/*
 * OOP class of RGTP connections
 * this file implements
 *   rgtp-open host port [username [accesslevel] md5secret]
 *      which returns the connection command name, and the
 *      per-connection command.  If username and md5secret
 *      aren't given then no subcommands which require anything
 *      except null access are allowed.  If they are then
 *      register isn't allowed.
 *
 * Failure codes (in errorCode) are:
 *   RGTP INT                  An operation was interrupted (eg, ^C)
 *   RGTP BADSERVER NOAUTH     The server didn't ask us for auth though we were
 *                              expecting it.  This may indicate a spoof server.
 *   RGTP BADSERVER BADAUTH    The server gave us bad crypto auth details. This
 *                              almost certainly indicates a spoof server.
 *   RGTP BADUSER              The server didn't like our userid (or access level).
 *   RGTP BADSECRET SYNTAX     The supplied MD5 secret was badly formatted.
 *   RGTP BADSECRET MISSING    The server wanted to do auth, but no secret was given.
 *   RGTP BADSECRET INCORRECT  The server didn't like our crypto auth - the MD5
 *                              secret must have been wrong.
 *
 *   rgtp-compareseqs value reference
 *      compares value against reference using wrap-around arithmetic
 *      returns 1 if value is larger, -1 if smaller or 0 if the same
 *
 * The sub-commands are in other files, except the management commands:
 *   $conn accesslevel [accesslevel] md5secret
 *   $conn close
 *   $conn destroy
 *   $conn noop
 *   $conn status [-access|-naccess] [-locked|-nlocked] \
 *                [-editing|-nediting] [-continuing|-ncontinuing]
 *      default is -access -locked -editing -continuing
 *      return is list containing items for:
 *       access:     broken|read|write|edit|register    naccess:     0|1|2|3|-1
 *       locked:     ""|locked                          nlocked:     0|1
 *       editing:    ""|editing                         nediting:    0|1
 *       continuing: ""|<contitemid>                    ncontinuing: 0|1
 *
 *   $conn rawcommandresponse command
 *      return is response, or error if we got 5xx
 *      if response indicates more lines of response or data to follow
 *      the script writer must use rawreadline or rawreaddata to get them
 *   $conn rawreadline
 *      must be called when server is sending us a response line
 *      return is the line we read
 *   $conn rawreaddata
 *      must be called when server is sending us .-terminated data
 *      return is the data we read
 *   $conn rawtransmitdata data
 *      return is OK, or list: FORMAT TEXT|SUBJECT|GROGNAME reason
 *
 *   $conn debug [-on|-off]
 *
 * ... not yet implemented:
 *   rgtp-open -register host port
 *      opens a connection and does a REGU operation
 *      the information text returned from REGU is in itemdata
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

static const RG_KeyValuePair rawdataerrors[] = {
  { "423 ", "FORMAT TEXT"     },
  { "424 ", "FORMAT SUBJECT"  },
  { "425 ", "FORMAT GROGNAME" },
  {  0                        }
};

int RG_Object_DebugCmd(Tcl_Interp *interp, RG_Connection *connP,
                       int argc, char **argv) {
  int olddebug= connP->debug;
  if (argc == 2) {
    if (!strcmp(argv[1],"-on")) connP->debug= 1;
    else if (!strcmp(argv[1],"-off")) connP->debug= 0;
    else {
      interp->result= "bad option to \"(conn) debug\" - must be -on or -off";
      return TCL_ERROR;
    }
  } else if (argc != 1) {
    interp->result= "wrong # args"; return TCL_ERROR;
  }
  sprintf(interp->result,"%d",olddebug);
  return TCL_OK;
}

static int RG_Status_AccessFn(Tcl_Interp *interp, RG_Connection *connP, int numeric) {
  int v;
  v= connP->status == RG_Status_Broken ? RG_Access_None
                                       : !connP->user ? -1 : connP->access;
  assert(v >= -1 && v <= RG_Access_Edit);
  return v;
}

static int RG_Status_LockedFn(Tcl_Interp *interp, RG_Connection *connP, int numeric) {
  assert(connP->editlocked >= 0 && connP->editlocked <= 1);
  return connP->editlocked;
}

static int RG_Status_EditingFn(Tcl_Interp *interp, RG_Connection *connP, int numeric) {
  assert(connP->editing >= 0 && connP->editing <= 1);
  return connP->editing;
}

static int RG_Status_ContFn(Tcl_Interp *interp, RG_Connection *connP, int numeric) {
  if (connP->contitem[0] && !numeric) {
    Tcl_AppendElement(interp,connP->contitem); return -1;
  }
  return connP->contitem[0] != 0;
}

int RG_Object_StatusCmd(Tcl_Interp *interp, RG_Connection *connP,
                           int argc, char **argv) {
  static struct {
    char *cmd;
    int (*fn)(Tcl_Interp*,RG_Connection*,int numflag);
    int base;
    char *words[5];
  } table[] = { 
    { "access",     RG_Status_AccessFn, -1, { "register","broken","read",
                                                "write","edit"            }},
    { "locked",     RG_Status_LockedFn,  0, { "","locked"                 }},
    { "editing",    RG_Status_EditingFn, 0, { "","editing"                }},
    { "continuing", RG_Status_ContFn,    0, { ""                          }},
    { 0 }
  }, *tp;
  
  int r;
  char *p;
  char buf[20];
  
  if (argc == 1) {
    for (tp= table; tp->cmd; tp++) {
      r= tp->fn(interp,connP,0);
      if (r >= 0) Tcl_AppendElement(interp, tp->words[r-tp->base]);
    }
  } else {
    while ((p= *++argv) != 0) {
      int n= -1;
      if (*p=='-') {
        for (tp= table; tp->cmd; tp++) {
          if (!strcmp(p+1,tp->cmd)) { n=0; break; }
          if (p[1]=='n' && !strcmp(p+2,tp->cmd)) { n=1; break; }
        }
      }
      if (n >= 0) {
        r= tp->fn(interp,connP,n);
        if (r >= 0) {
          if (n) { sprintf(buf,"%d",r); p=buf; }
          else { p= tp->words[r-tp->base]; }
          Tcl_AppendElement(interp, p);
        }
      } else {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp,
                         "unknown RGTP status option \"",
                         p, "\": must be one of", (char*)0);
        for (tp= table; tp->cmd; tp++)
          Tcl_AppendResult(interp, " -[n]",tp->cmd,(char*)0);
        return TCL_ERROR;
      }
    }
  }
  return TCL_OK;
}

int RG_Object_AccessLevelCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv) {
  int v, r;
  int oldmd5secretbytes;
  char buf[10];
  Tcl_DString ds;
  unsigned char oldmd5secret[SECRET_MAXBYTES];

  if (argc != 2 && argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (!connP->user) {
    interp->result= "can't set access level for registration connection";
    return TCL_ERROR;
  }

  r= RG_CommandResponseRetry(interp,connP,&ds,"NOOP");
  if (r != TCL_OK) return r;
  if (strncmp(Tcl_DStringValue(&ds),"200 ",4)) {
    Tcl_AppendResult(interp,
                     "server refused our NOOP before access level change:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  
  oldmd5secretbytes= connP->md5secretbytes;
  memcpy(oldmd5secret,connP->md5secret,oldmd5secretbytes);
  if (argc == 3) {
    r= RG_GetAccessLevel(interp,argv[1],&v);  if (r != TCL_OK) return r;
    r= RG_GetMD5Secret(interp,connP,argv[2]);
    sprintf(buf," %d",v);
  } else {
    r= RG_GetMD5Secret(interp,connP,argv[1]);  if (r != TCL_OK) return r;
    v= -1; buf[0]= 0;
  }
  r= RG_CommandResponse(interp,connP,&ds,"ALVL%s",buf);  if (r != TCL_OK) return r;
  r= RG_HandleAuth(interp,connP,&ds,v,0);
  if (r == TCL_OK) {
    connP->wantaccess= v;
  } else {
    connP->md5secretbytes= oldmd5secretbytes;
    memcpy(connP->md5secret,oldmd5secret,oldmd5secretbytes);
  }
  return r;
}

int RG_Object_RegisterCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  Tcl_DString ds;
  int r;
  char *cp, *sp;
  unsigned long l;

  if (argc != 1 && argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (connP->user) { interp->result= "conn with user unregistrable"; return TCL_ERROR; }

  if (argc == 2) {
    r= RG_CommandResponseRetry(interp,connP,&ds,"NOOP");
    if (r != TCL_OK) return r;
    if (strncmp(Tcl_DStringValue(&ds),"200 ",4)) {
      Tcl_AppendResult(interp,
                       "server refused our NOOP before registration confirmation:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
  } else if (connP->regudone) {
    assert(argc==1);
    interp->result= "register pt.1 already done"; return TCL_ERROR;
  }
  if (!connP->regudone) {
    r= RG_CommandResponse(interp,connP,&ds,"REGU"); if (r != TCL_OK) return r;
    if (strncmp(Tcl_DStringValue(&ds),"100 ",4)) {
      Tcl_AppendResult(interp,
                       "server inform us it was starting registration:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    r= RG_ReadLineEcho(interp, connP, &ds); if (r != TCL_OK) return r;
    if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
      Tcl_AppendResult(interp,
                       "server didn't allow us to start registration:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    r= RG_ReadData(interp,connP,&ds);
    if (r != TCL_OK) return r;
    connP->regudone= 1;
    if (argc == 1) {
      cp= Tcl_DStringValue(&ds);
      for (;;) {
        sp= strchr(cp,'\n');
        if (*cp == ' ') {
          if (sp) *sp= 0;
          l= sp ? (long)(sp-cp) : strlen(cp);
          Tcl_AppendResult(interp,cp,"\n",(char*)0);
        }
        if (!sp) break;
        cp= sp+1;
      }
      Tcl_DStringFree(&ds);
      return TCL_OK;
    } else {
      Tcl_DStringFree(&ds);
    }
  }
  assert(argc==2);
  r= RG_CommandResponse(interp,connP,&ds,"USER %s",argv[1]);
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"482 ",4)) {
    Tcl_AppendResult(interp,
                     "server rejected our registration:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_SetErrorCode(interp, "RGTP","BADUSER",Tcl_DStringValue(&ds),(char*)0);
    Tcl_DStringFree(&ds);
    connP->status= RG_Status_NeedReconnect;
    return TCL_ERROR;
  } else if (!strncmp(Tcl_DStringValue(&ds),"280 ",4)) {
    Tcl_AppendResult(interp,Tcl_DStringValue(&ds)+4,(char*)0);
    Tcl_DStringFree(&ds);
    RG_Break(connP);
    return TCL_OK;
  } else {
    Tcl_AppendResult(interp,
                     "server gave bad response to our registration request:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
}

int RG_Object_NoOpCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  Tcl_DString ds;
  int r;

  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_None) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds, "NOOP");
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"200 ",4)) { Tcl_DStringFree(&ds); return TCL_OK; }
  Tcl_AppendResult(interp,
                   "NOOP command rejected by server:\n",
                   Tcl_DStringValue(&ds), (char*)0);
  Tcl_DStringFree(&ds);
  return TCL_ERROR;
}

int RG_Object_RawCommandResponseCmd(Tcl_Interp *interp, RG_Connection *connP,
                                    int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (strlen(argv[1]) > TXRXLINE_MAXLEN || strchr(argv[1],'\n')) {
    interp->result= "bad format for quoted line"; return TCL_ERROR;
  }
  if (connP->status != RG_Status_OK) { interp->result= "bad status"; return TCL_ERROR; }
  r= RG_CommandResponse(interp,connP,&ds,"%s",argv[1]); if (r != TCL_OK) return r;
  Tcl_DStringResult(interp,&ds);
  return TCL_OK;
}

int RG_Object_RawReadLineCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (connP->status != RG_Status_OK) { interp->result= "bad status"; return TCL_ERROR; }
  r= RG_ReadLineEcho(interp, connP, &ds); if (r != TCL_OK) return r;
  Tcl_DStringResult(interp,&ds);
  return TCL_OK;
}

int RG_Object_RawReadDataCmd(Tcl_Interp *interp, RG_Connection *connP,
                             int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (connP->status != RG_Status_OK) { interp->result= "bad status"; return TCL_ERROR; }
  r= RG_ReadData(interp, connP, &ds); if (r != TCL_OK) return r;
  Tcl_DStringResult(interp,&ds);
  return TCL_OK;
}

int RG_Object_RawTransmitDataCmd(Tcl_Interp *interp, RG_Connection *connP,
                                 int argc, char **argv) {
  int r;
  
  if (argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (connP->status != RG_Status_OK) { interp->result= "bad status"; return TCL_ERROR; }
  r= RG_TransmitData(interp, connP, argv[1],argv[2], rawdataerrors);
  if (r == TCL_RETURN) return TCL_OK;
  if (r != TCL_OK) return r;
  interp->result= "OK";
  return TCL_OK;
}

int RG_Object_CloseCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  Tcl_DString ds;
  int r;

  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (connP->status == RG_Status_OK) {
    r= RG_CommandResponse(interp,connP,&ds, "QUIT");
  } else {
    r= TCL_ERROR;
  }
  RG_DeleteCommand(interp,connP);
  if (r != TCL_OK) return TCL_OK;
  if (strncmp(Tcl_DStringValue(&ds),"280 ",4) &&
      strncmp(Tcl_DStringValue(&ds),"481 ",4)) {
    Tcl_DStringResult(interp,&ds);
  } else {
    Tcl_DStringFree(&ds);
  }
  return r;
}

int RG_Object_DestroyCmd(Tcl_Interp *interp, RG_Connection *connP,
                         int argc, char **argv) {
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  RG_DeleteCommand(interp,connP);
  return TCL_OK;
}

int RG_ConnectionCmd(ClientData clientData, Tcl_Interp *interp,
                     int argc, char **argv) {
  static const struct {
    const char *cmd;
    int (*fn)(Tcl_Interp*,RG_Connection*,int,char**);
  } table[] = {
    /* Status and general commands */
    { "close",              RG_Object_CloseCmd              },
    { "destroy",            RG_Object_DestroyCmd            },
    { "noop",               RG_Object_NoOpCmd               },
    { "status",             RG_Object_StatusCmd             },
    { "debug",              RG_Object_DebugCmd              },
    { "accesslevel",        RG_Object_AccessLevelCmd        },
    { "register",           RG_Object_RegisterCmd           },

    /* Reading commands */
    { "index",              RG_Object_IndexCmd              },
    { "indexinfo",          RG_Object_IndexInfoCmd          },
    { "indexlines",         RG_Object_IndexLinesCmd         },
    { "indexentry",         RG_Object_IndexEntryCmd         },
    { "indexmunge",         RG_Object_IndexModifyCmd        },
    { "item",               RG_Object_ItemCmd               },
    { "itemmark",           RG_Object_ItemMarkCmd           },
    { "itemdata",           RG_Object_ItemDataCmd           },
    { "itemstat",           RG_Object_ItemStatCmd           },
    { "motd",               RG_Object_MoTDCmd               },
    { "editlog",            RG_Object_EditLogCmd            },

    /* Writing commands */
    { "reply",              RG_Object_ReplyCmd              },
    { "continue",           RG_Object_ContinueCmd           },
    { "contribute",         RG_Object_ContributeCmd         },

    /* Editor commands */
    { "editlock",           RG_Object_EditLockCmd           },
    { "editunlock",         RG_Object_EditUnlockCmd         },
    { "edit",               RG_Object_EditCmd               },
    { "editdone",           RG_Object_EditDoneCmd           },
    { "editindex",          RG_Object_EditIndexCmd          },
    { "withdraw",           RG_Object_WithdrawCmd           },
    { "setmotd",            RG_Object_SetMoTDCmd            },
    { "udbm",               RG_Object_UDBMCmd               },
    { "killserver",         RG_Object_KillServerCmd         },
    { "restartserver",      RG_Object_RestartServerCmd      },

    /* Raw protocol access commands */
    { "rawcommandresponse", RG_Object_RawCommandResponseCmd },
    { "rawreadline",        RG_Object_RawReadLineCmd        },
    { "rawreaddata",        RG_Object_RawReadDataCmd        },
    { "rawtransmitdata",    RG_Object_RawTransmitDataCmd    },
    { 0 }
  }, *tp;
  RG_Connection *connP= (RG_Connection*)clientData;
  int r;

  if (argc < 2) {
    interp->result= "insufficient # args - need subcommand";
    return TCL_ERROR;
  }

  for (tp= table;
       tp->cmd && strcmp(argv[1],tp->cmd);
       tp++);
  if (tp->cmd) {
    r= tp->fn(interp,connP,--argc,++argv);
    return r;
  } else {
    Tcl_AppendResult(interp,
                     "unknown RGTP command \"",
                     argv[1],
                     "\": must be one of",
                     (char*)0);
    for (tp= table; tp->cmd; tp++)
      Tcl_AppendResult(interp, " ",tp->cmd,(char*)0);
    return TCL_ERROR;
  }
}

int RG_OpenCmd(ClientData clientData, Tcl_Interp *interp,
               int argc, char **argv) {
  static int id= 0;
  RG_Connection *connP;
  int r, v;
  int debug= 0;
  long l;
  char *p;
  struct servent *serv;

  if (argv[1] && !strcmp(argv[1],"-debug")) { debug=1; argc--; argv++; }

  if (argc != 3 && argc != 5 && argc != 6) {
    interp->result= "wrong # args"; return TCL_ERROR;
  }
  connP= (RG_Connection*)ckalloc(sizeof(RG_Connection));
  Tcl_DStringInit(&connP->itemtext);
  connP->status= RG_Status_Broken;
  connP->debug= debug;
  connP->id= id++;
  RG_EArrayInit(&connP->indexdata);
  RG_EArrayInit(&connP->indexorder);
  RG_EArrayInit(&connP->marks);
  connP->indexsince= 0UL;
  connP->itemvalid= connP->indexvalid= connP->editing= connP->editlocked= 0;
  connP->regudone= connP->logindone= 0;
  connP->contitem[0]= 0;

  if (!argv[2][0]) {
    connP->port= htons(TCPPORT_RGTP);
  } else {
    l= strtol(argv[2],&p,10);
    if (!*p) {
      connP->port= htons(l);
    } else {
      serv= getservbyname(argv[2],"tcp");
      endservent();
      if (!serv) {
        ckfree(connP);
        Tcl_AppendResult(interp,
                         "unknown service name \"", argv[2],
                         "\": should be number or protocol name",
                         (char*)0);
        return TCL_ERROR;
      }
      connP->port= serv->s_port;
    }
  }

  connP->host= (char*)ckalloc(strlen(argv[1])+1);
  strcpy(connP->host, argv[1]);

  if (argc >= 5) {
    r= RG_GetMD5Secret(interp,connP,argv[argc-1]);
    if (r != TCL_OK) { ckfree(connP->host); ckfree(connP); return r; }
    connP->user= (char*)ckalloc(strlen(argv[argc-2])+1);
    strcpy(connP->user, argv[argc-2]);

    if (argc == 6) {
      r= RG_GetAccessLevel(interp,argv[3],&v);
      connP->wantaccess= v;
      if (r != TCL_OK) return r;
    } else {
      connP->wantaccess= -1;
    }
  } else {
    connP->md5secretbytes= 0;
    connP->user= 0;
  }
  if (connP->debug) fprintf(stderr,"RGTP %d (init)\n",connP->id);

  r= RG_Reconnect(interp,connP);
  if (r != TCL_OK) {
    ckfree(connP->host);
    ckfree(connP->user);
    ckfree(connP);
    return r;
  }
  p= RG_CommandName(connP);
  Tcl_CreateCommand(interp, p, RG_ConnectionCmd,
                    (ClientData)connP, RG_DeleteCallback);
  strcpy(interp->result, p);
  if (connP->debug) fprintf(stderr,"RGTP %d (created)\n",connP->id);
  return TCL_OK;
}

int RG_CompareSeqsCmd(ClientData clientData, Tcl_Interp *interp,
                      int argc, char **argv) {
  unsigned long a, b;
  char *q;
  int compres;
 
  if (argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }

  a= strtoul(argv[1], &q, 16);
  if (!q) { interp->result= "compareseqs val arg broken"; return TCL_ERROR; }
  b= strtoul(argv[2], &q, 16);
  if (!q) { interp->result= "compareseqs ref arg broken"; return TCL_ERROR; }

  compres= RG_Compare_WrappingLongs(a,b);
  sprintf(interp->result,"%d",compres);
  return TCL_OK;
}

int RGTP_TclInit(Tcl_Interp *interp) {
  Tcl_CreateCommand(interp, "rgtp-open", RG_OpenCmd,
                    (ClientData)0, (void(*)())0);
  Tcl_CreateCommand(interp, "rgtp-compareseqs", RG_CompareSeqsCmd,
                    (ClientData)0, (void(*)())0);
  return TCL_OK;
}
