/*
 * OOP class of RGTP connections
 * implements reading commands (except index ones)
 *   $conn item itemid
 *      return is continued-from, continued-in, last-edit, last-reply
 *   $conn itemstat itemid
 *      return is continued-from, continued-in, last-edit, last-reply, subject
 *   $conn itemmark [-<reqop>-<field> <value>]
 *                  [-<wantop>-<field>]
 *                  [-<field> ...]
 *     <reqop> is one of    gt lt ge le
 *     <wantop> is one of   first last
 *     <field> is one of    index line seq time
 *     <value> should be 8 hex digits for seq or time,
 *       or a decimal number for index or line
 *     return is list, either  <index> <line> <seq> <time>
 *       or items specified in order
 *     if no -reqop then return is list of the things for each item
 *   $conn itemdata
 *   $conn editlog
 *   $conn motd
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

int RG_Object_MoTDCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  Tcl_DString ds;
  int r;
  char *ep, *eq, *sp, *ss;
  unsigned long timestamp, cookie=0;
  unsigned long l;

  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_None) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds,"MOTD");
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"410 ",4)) {
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide message of the day:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  sp= Tcl_DStringValue(&ds);
  timestamp= strtoul(sp,&ep,16);
  if (ep == sp+8 && *ep == ' ') cookie= strtoul(sp+9,&eq,16);
  else eq= 0;
  if (eq != sp+17 || *eq != '\n') {
    Tcl_AppendResult(interp, "server sent bad motd format:\n", sp, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  sprintf(interp->result,"%08lX %08lX",cookie,timestamp);
  ss= ep= ++eq;
  do {
    if (*ep == '^') ep++;
    sp= strchr(ep,'\n');
    l= sp ? sp-ep+1 : strlen(ep)+1;
    if (eq!=ep) { /* I'd like to use memmove, but ScumOS doesn't have it */
      while (l--) *eq++= *ep++;
    } else {
      ep +=l; eq += l;
    }
  } while (sp);
    
  Tcl_AppendElement(interp,ss);
  Tcl_DStringFree(&ds);
  return TCL_OK;
}

int RG_Object_EditLogCmd(Tcl_Interp *interp, RG_Connection *connP,
                         int argc, char **argv) {
  Tcl_DString ds;
  int r;
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Read) != TCL_OK) return TCL_ERROR;
  r= RG_CommandResponseRetry(interp,connP,&ds,"ELOG");
  if (r != TCL_OK) return r;
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp, "server didn't provide edit log:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  Tcl_DStringResult(interp,&ds);
  return TCL_OK;
}

static int encap8orspaces(Tcl_Interp *interp, char *p) {
  int i;
  char *q;
  if (*p == ' ') {
    if (strspn(p," ") != 8) return TCL_ERROR;
    Tcl_AppendElement(interp,"");
  } else {
    for (i=8, q=p; i--; q++) if (!isalnum(*q)) return TCL_ERROR;
    Tcl_AppendElement(interp,p);
  }
  return TCL_OK;
}

int RG_Compare_WrappingLongs(unsigned long val, unsigned long ref) {
  unsigned long halfway = (~0UL) >> 1;
  unsigned long difference= val - ref;
  return difference == 0UL     ?  0
    :    difference <= halfway ?  1
    :                            -1;
}

enum { rl_gt, rl_lt, rl_ge, rl_le };
enum { mff_index, mff_line, mff_seq, mff_time };

static unsigned long markvalue(RG_MarkInItem *marks, int i, int key) {
  switch (key) {
  case mff_index:   return i;
  case mff_line:    return marks[i].linenumber;
  case mff_seq:     return marks[i].cookie;
  case mff_time:    return marks[i].timestamp;
  }
  abort();
}

enum RG_Relations { RG_Rel_GT, RG_Rel_LT, RG_Rel_GE, RG_Rel_LE };

static int marksatisfactory(RG_MarkInItem *marks, int i,
                            int sense, int key, unsigned long refvalue) {
  int compres;
  
  if (sense == -1) return 1;
  compres= RG_Compare_WrappingLongs(markvalue(marks,i,key), refvalue);
  
  switch (sense) {
  case RG_Rel_GT:  return compres >  0;
  case RG_Rel_LT:  return compres <  0;
  case RG_Rel_GE:  return compres >= 0;
  case RG_Rel_LE:  return compres <= 0;
  }
  abort();
}

static void markfill(Tcl_DString *dsP, RG_MarkInItem *marks, int i, int *fields) {
  unsigned long v;
  char buf[30];
  
  Tcl_DStringInit(dsP);
  while (*fields != -1) {
    v= markvalue(marks,i,*fields);
    sprintf(buf, *fields <= mff_line ? "%lu" : "%08lx", v);
    Tcl_DStringAppendElement(dsP,buf);
    fields++;
  }
}

int RG_Object_ItemMarkCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  /*
   *   $conn itemmark [-<reqop>-<field> value]
   *                  [-<wantop>-<field>]
   *                  [-<field> ...]
   *     <reqop> is one of    gt lt ge le
   *     <wantop> is one of   first last
   *     <field> is one of    index line seq time
   *     return is list, either  <index> <line> <seq> <time>
   *       or items specified in order
   *     if no -reqop then return is list of the things for each item
   */
  static char *fieldnames[]= { "index", "line", "seq", "time", 0 };
  static char *relationnames[]= { "gt-", "lt-", "ge-", "le-", 0 };
  /* relationnames must be in the same order as RG_Relations */

  int fieldslist[5];
  int fieldswanted= 0;
  int bestsense= 0; /* -1=first, 1=last, 0=all */
  int bestkey= -1; /* -1 only if bestsense==0 */
  int restrictkey= -1; /* -1 = no restriction */
  int restrictsense= -1; /* -1 = no restriction */
  unsigned long resrefval= 0;
  char *p, **pp, *q;
  int i, j;
  Tcl_DString ds;
  RG_MarkInItem *marks;
  
  if (!connP->itemvalid) {
    interp->result= "no item marks stored - use the item command";
    return TCL_ERROR;
  }

  while ((p= *++argv) != 0) {
    if (*p++ != '-') { interp->result= "mark must take options"; return TCL_ERROR; }
    q= strchr(p,'-');
    if (!q) { q=p; p=0; } else { q++; }
    for (pp=fieldnames, i=0; *pp && strcmp(q,*pp); pp++, i++);
    if (!*pp) { interp->result= "unknown mark field"; return TCL_ERROR; }
    if (p) {
      if (!strncmp(p,"first-",6)) { bestkey= i; bestsense= -1; continue; }
      if (!strncmp(p,"last-",5)) { bestkey= i; bestsense= 1; continue; }
      for (pp=relationnames, j=0; *pp && strncmp(p,*pp,3); pp++, j++);
      if (!*pp) { interp->result= "unknown mark field preceder"; return TCL_ERROR; }
      restrictkey= i; restrictsense= j;
      if (!(p= *++argv)) { interp->result= "mark no restrict value"; return TCL_ERROR; }
      resrefval= strtoul(p, &q, restrictkey >= mff_seq ? 16 : 10);
      if (*q) { interp->result= "mark bad restrict value"; return TCL_ERROR; }
    } else {
      fieldslist[fieldswanted++]= i;
      if (fieldswanted >= sizeof(fieldslist)/sizeof(int)) {
        interp->result= "too many mark fields wanted"; return TCL_ERROR;
      }
    }
  }
  if (fieldswanted == 0) {
    while (fieldswanted <= mff_time) {
      fieldslist[fieldswanted]= fieldswanted;
      fieldswanted++;
    }
  }
  fieldslist[fieldswanted]= -1;

  marks= (RG_MarkInItem*)connP->marks.data;
  if (bestsense != 0) {
    if (bestkey <= mff_line) {
      if (bestsense > 0) {
        for (i= connP->marks.used-1;
             i>=0 && !marksatisfactory(marks,i,restrictsense,restrictkey,resrefval);
             i--);
      } else {
        for (i= 0;
             i<connP->marks.used &&
             !marksatisfactory(marks,i,restrictsense,restrictkey,resrefval);
             i++);
        if (i >= connP->marks.used) i=-1;
      }
    } else {
      i= -1;
      for (j= connP->marks.used-1;
           j>=0 && !marksatisfactory(marks,j,restrictsense,restrictkey,resrefval);
           j--)
        if (i==-1 ||
            (bestsense > 0 ? marks[j].cookie > marks[i].cookie :
                             marks[j].cookie < marks[i].cookie))
          i= j;
    }
    if (i >= 0) {
      markfill(&ds,marks,i,fieldslist);
      Tcl_DStringResult(interp,&ds);
    }
  } else {
    for (i= 0;
         i<connP->marks.used;
         i++)
      if (marksatisfactory(marks,i,restrictsense,restrictkey,resrefval)) {
        markfill(&ds,marks,i,fieldslist);
        Tcl_AppendElement(interp,Tcl_DStringValue(&ds));
        Tcl_DStringFree(&ds);
      }
  }
  return TCL_OK;
}

int RG_Object_ItemDataCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (!connP->itemvalid) {
    interp->result= "no item text stored - use the item command";
    return TCL_ERROR;
  }
  Tcl_SetResult(interp,Tcl_DStringValue(&connP->itemtext),TCL_VOLATILE);
  return TCL_OK;
}

int RG_Object_ItemCmd(Tcl_Interp *interp, RG_Connection *connP,
                      int argc, char **argv) {
  char buf[20], *p, *q, *ep, *sp;
  Tcl_DString ds;
  int r, linenumber;
  RG_MarkInItem *mark;

  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Read) != TCL_OK) return TCL_ERROR;
  if (strlen(argv[1]) != ITEMID_LEN) {
    interp->result= "item-id is wrong length"; return TCL_ERROR;
  }
  strcpy(buf,"ITEM "); strcat(buf,argv[1]);

  r= RG_CommandResponseRetry(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  if (!strncmp(Tcl_DStringValue(&ds),"410 ",4)) {
    Tcl_DStringFree(&connP->itemtext);
    Tcl_DStringAppend(&connP->itemtext,Tcl_DStringValue(&ds)+3,-1);
    Tcl_DStringFree(&ds);
    connP->itemvalid= 1;
    return TCL_OK;
  }
  if (strncmp(Tcl_DStringValue(&ds),"250 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide item ", argv[1],
                     " on request:\n", Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  r= RG_ReadData(interp,connP,&ds);
  if (r != TCL_OK) return r;
  sp= Tcl_DStringValue(&ds);
  p= strchr(sp,'\n');
  if (p != sp + 35 || sp[8] != ' ' ||
      sp[17] != ' ' || sp[26] != ' ') {
    Tcl_AppendResult(interp,
                     "server provided bad status line for item ", argv[1],
                     " - item data:\n", sp, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  sp[8]= sp[17]= sp[26]= *p++= 0;
  connP->itemvalid= 0;
  Tcl_DStringFree(&connP->itemtext);
  RG_EArrayReset(&connP->marks);
  linenumber= 0;
  while (*p) {
    q= strchr(p,'\n');
    assert(q);
    *q++= 0;
    if (*p == '^' && *++p !='^' ) {
      mark= (RG_MarkInItem*) RG_EArrayAppend(&connP->marks,
                                             sizeof(RG_MarkInItem));
      if (q != p+18 || p[8] != ' ') {
        ep= 0;
      } else {
        mark->linenumber= linenumber;
        mark->cookie= strtoul(p,&ep,16);
        if (ep != p+8) {
          ep= 0;
        } else {
          mark->timestamp= strtoul(p+9,&ep,16);
          if (ep != p+17) ep= 0;
        }
      } 

      if (!ep) {
        Tcl_AppendResult(interp,
                         "server provided bad cookie line in item ", argv[1],
                         " - cookie was:\n", --p, (char*)0);
        Tcl_DStringFree(&ds);
        return TCL_ERROR;
      }
    } else {
      Tcl_DStringAppend(&connP->itemtext,p,-1);
      Tcl_DStringAppend(&connP->itemtext,"\n",1);
      linenumber++;
    }
    p= q;
  }
  r= encap8orspaces(interp,sp);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+9);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+18);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+27);
  if (r!=TCL_OK) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,
                     "server sent bad status line for item ", argv[1],
                     ":\n", sp, (char*)0);
    RG_Violation(connP);
    Tcl_DStringFree(&ds);
    return r;
  }
  Tcl_DStringFree(&ds);
  connP->itemvalid= 1;
  return TCL_OK;
}

int RG_Object_ItemStatCmd(Tcl_Interp *interp, RG_Connection *connP,
                          int argc, char **argv) {
  char buf[20], *sp, *subjp;
  Tcl_DString ds;
  int r, l;
  
  if (argc != 2) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (RG_AccessOK(interp,connP,RG_Access_Read) != TCL_OK) return TCL_ERROR;
  if (strlen(argv[1]) != ITEMID_LEN) {
    interp->result= "item-id is wrong length"; return TCL_ERROR;
  }
  strcpy(buf,"STAT "); strcat(buf,argv[1]);

  r= RG_CommandResponseRetry(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  sp= Tcl_DStringValue(&ds);
  if (!strncmp(sp,"410 ",4)) {
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(sp,"211 ",4)) {
    Tcl_AppendResult(interp,
                     "server didn't provide item info for ", argv[1],
                     " on request:\n", Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  sp+= 4;
  l= strlen(sp);
  subjp= sp+35;
  if (l < 35 || sp[8] != ' ' || sp[17] != ' ' || sp[26] != ' ' ||
      (*subjp && *subjp != ' ')) {
    Tcl_AppendResult(interp,
                     "server provided bad status for item ", argv[1],
                     ":\n", sp-4, (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  sp[8]= sp[17]= sp[26]= 0;
  if (*subjp) *subjp++= 0;
  r= encap8orspaces(interp,sp);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+9);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+18);
  if (r==TCL_OK) r= encap8orspaces(interp,sp+27);
  if (r!=TCL_OK) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,
                     "server sent bad status for item ", argv[1],
                     ":\n", sp-4, (char*)0);
    Tcl_DStringFree(&ds);
    return r;
  }
  Tcl_AppendElement(interp,subjp);
  Tcl_DStringFree(&ds);
  return TCL_OK;
}

