/*
 * Assistance functions for tty-based program
 * this file implements
 *   tty size widthvar heightvar
 *   tty stringmult base len
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <termios.h>

#include "tcl.h"
#include "tclExtend.h"

#include "fixup.h"
#include "Tty.h"

int Tty_StringMultCmd(Tcl_Interp *interp, int argc, char **argv) {
  int len, lenleft, each, r;
  char *build, *at;
  
  if (argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }
  each= strlen(argv[1]);
  if (!each) { interp->result= "stringmult empty base"; return TCL_ERROR; }
  r= Tcl_GetInt(interp,argv[2],&len); if (r != TCL_OK) return r;
  if (len<0) { interp->result= "stringmult negative len"; return TCL_ERROR; }

  build= ckalloc(len+1);
  for (lenleft=len, at=build; lenleft >= each; at+=each, lenleft-=each)
    memcpy(at,argv[1],each);
  if (lenleft) {
    memcpy(at,argv[1],lenleft); at+=lenleft;
  }
  *at= 0;
  Tcl_SetResult(interp,build,TCL_DYNAMIC);
  return TCL_OK;
}

int Tty_SetRestartSIGWINCHCmd(Tcl_Interp *interp, int argc, char **argv) {
#ifdef SA_RESTART
  struct sigaction sa;
  char *p;
#endif
  
  if (argc != 1) { interp->result= "wrong # args"; return TCL_ERROR; }
#ifdef SA_RESTART
  if (sigaction(SIGWINCH,0,&sa)) {
    p= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't get old SIGWINCH action: ", p, (char*)0);
    return TCL_ERROR;
  }
  sa.sa_flags |= SA_RESTART;
  if (sigaction(SIGWINCH,&sa,0)) {
    p= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't set new SIGWINCH action: ", p, (char*)0);
    return TCL_ERROR;
  }
#endif
  return TCL_OK;
}

int Tty_SizeCmd(Tcl_Interp *interp, int argc, char **argv) {
  struct winsize fromioctl;
  int lines, columns;
  char buf[40], *e;

  lines= TTY_DEFAULTLINES;
  columns= TTY_DEFAULTCOLUMNS;
  if ((e= getenv("LINES"))!=0) { lines= atoi(e); }
  if ((e= getenv("COLUMNS"))!=0) { columns= atoi(e); }
  if (!ioctl(0,TIOCGWINSZ,(void*)&fromioctl)) {
    if (fromioctl.ws_row) lines= fromioctl.ws_row;
    if (fromioctl.ws_col) columns= fromioctl.ws_col;
  }

  if (argc == 2) {
    if (!strcmp(argv[1],"-width")) {
      sprintf(interp->result,"%d",lines);
    } else if (!strcmp(argv[1],"-height")) {
      sprintf(interp->result,"%d",columns);
    } else {
      interp->result= "tty size single arg must be -width or -height";
      return TCL_ERROR;
    }
    return TCL_OK;
  } else if (argc > 3) {
    interp->result= "wrong # args"; return TCL_ERROR;
  }
  sprintf(buf,"%d",columns);
  Tcl_AppendElement(interp,buf);
  if (argc == 3)
    if (!Tcl_SetVar(interp,argv[1],buf,TCL_LEAVE_ERR_MSG)) return TCL_ERROR;

  sprintf(buf,"%d",lines);
  Tcl_AppendElement(interp,buf);
  if (argc == 3)
    if (!Tcl_SetVar(interp,argv[2],buf,TCL_LEAVE_ERR_MSG)) return TCL_ERROR;

  return TCL_OK;
}

int Tty_Cmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv) {
  static const struct {
    const char *cmd;
    int (*fn)(Tcl_Interp*,int,char**);
  } table[] = {
    { "size",                   Tty_SizeCmd                   },
    { "stringmult",             Tty_StringMultCmd             },
    { "setrestart-sigwinch",    Tty_SetRestartSIGWINCHCmd     },
    { 0 }
  }, *tp;

  if (argc < 2) {
    interp->result= "insufficient # args - need subcommand";
    return TCL_ERROR;
  }

  for (tp= table;
       tp->cmd && strcmp(argv[1],tp->cmd);
       tp++);
  if (tp->cmd) {
    return tp->fn(interp,--argc,++argv);
  } else {
    Tcl_AppendResult(interp,
                     "unknown tty command \"",
                     argv[1],
                     "\": must be one of",
                     (char*)0);
    for (tp= table; tp->cmd; tp++)
      Tcl_AppendResult(interp, " ",tp->cmd,(char*)0);
    return TCL_ERROR;
  }
}

int TtyAssist_TclInit(Tcl_Interp *interp) {
  Tcl_CreateCommand(interp, "tty", Tty_Cmd, (ClientData)0, (void(*)())0);
  return TCL_OK;
}
