/*
 * OOP class of .groggsrc files
 * implements
 *   groggsrc-listservers directory warnings lists
 *      which returns a list of lists, each list being
 *         hostname port nickname
 *      the first entry of the list of lists is the default
 *   groggsrc-server directory warnings hostname port
 *      which returns the server .groggsrc command name, and the
 *      per-server commands ($conn = server command):
 *   $server server [-hostname] [-port] [-nickname]
 *   $server login [-user] [-md5-secret]
 *   $server grogname     =>  grogname if any, otherwise getenv(), otherwise ""
 *   $server program [-editor] [-visual] [-default-editor] [-x11-editor] [-pager]
 *   $server seen-all-before [-set newvalue]
 *   $server iteminfo itemid [-junked] [-lastseen] \
 *                           [-setjunked 0|1|2] [-setlastseen newvalue]
 * Any warnings about oddities in the setup files are appended to $warnings
 */

#include "tcl.h"
#include "tclExtend.h"

#include "rgtp.h"

#ifdef BROKEN_STRTOUL
#define strtol(a,b,c) ((unsigned long)strtoul((a),(b),(c)))
#endif

char *serveridnames= {
  "hostname", "port", "nickname", 0
};
char *stringnames= {
  "grogname", "editor", "visual", "default-editor", "x11-editor", "pager", 0
};

char *junklevelnames= {
  "", "junk", "kill", 0
};

typedef struct {
  int junklevel;
  unsigned long lastseen;
  int lastseenvalid= 0;
} Grc_ItemInfo;

typedef struct {
  int id;
  char userid[USERID_MAXLEN];
  unsigned long md5secret;
  int md5secretvalid;
  Tcl_DString serverids[3];
  Tcl_DString strings[6];
  unsigned long seenallbefore;
  int seenallbeforevalid;
  Tcl_HashTable items;
} Grc_Server;

int RG_ConnectionCmd(ClientData clientData, Tcl_Interp *interp,
                     int argc, char **argv) {
  static const struct {
    const char *cmd;
    int (*fn)(Tcl_Interp*,RG_Connection*,int,char**);
  } table[] = {
    "close",         RG_Object_CloseCmd,
    "destroy",       RG_Object_DestroyCmd,
    "index",         RG_Object_IndexCmd,
    "indexinfo",     RG_Object_IndexInfoCmd,
    "indexlines",    RG_Object_IndexLinesCmd,
    "indexentry",    RG_Object_IndexEntryCmd,
    "item",          RG_Object_ItemCmd,
    "itemdata",      RG_Object_ItemDataCmd,
    "motd",          RG_Object_MoTDCmd,
    "noop",          RG_Object_NoOpCmd,
    "status",        RG_Object_StatusCmd,
    0
  }, *tp;
  RG_Connection *connP= (RG_Connection*)clientData;
  int r;
  
  if (argc < 2) {
    interp->result= "insufficient # args - need subcommand";
    return TCL_ERROR;
  }

  for (tp= table;
       tp->cmd && strcmp(argv[1],tp->cmd);
       tp++);
  if (tp->cmd) {
    r= tp->fn(interp,connP,--argc,++argv);
    return r;
  } else {
    Tcl_AppendResult(interp,
                     "unknown RGTP command \"",
                     argv[1],
                     "\": must be one of",
                     (char*)0);
    for (tp= table; tp->cmd; tp++)
      Tcl_AppendResult(interp, " ",tp->cmd,(char*)0);
    return TCL_ERROR;
  }
}

int Grc_OpenFiles(Tcl_Interp *interp, char *directory, char *warningsvar
                  FILE **rcfileP, FILE **secretfileP, int createguest) {
  char *bp, *msg;
  int l;
  struct stat stab;

  l= strlen(directory);
  if (!l) {
    directory= getenv("HOME");
    if (!directory) {
      interp->result= "No HOME environment variable - "
        "cannot find your home directory.";
      return TCL_ERROR;
    }
    l= strlen(directory);
  }
  bp= ckalloc(l+20);
  strcpy(bp,directory);
  while (l && bp[l-1]=='/') --l;
  bp[l++]= '/';

  strcpy(bp+l, ".groggsrc");
  *rcfileP= fopen(bp,"r");
  if (!*rcfileP) {
    if (errno == ENOENT) {
      if (createguest) {
        *rcfileP= fopen(bp,"w");
        if (!*rcfileP) {
          msg= Tcl_PosixError(interp);
          Tcl_AppendResult(interp, "couldn't create a .groggsrc for you: ", msg,
                           (char*)0);
          ckfree(bp); return TCL_ERROR;
        }
        if (!fputs("default-server\nuser GUEST\n",*rcfileP)) {
          msg= Tcl_PosixError(interp);
          Tcl_AppendResult(interp, "couldn't write into newly-created .groggsrc: ",
                           msg, (char*)0);
          fclose(*rcfileP);
          ckfree(bp); return TCL_ERROR;
        }
        if (fclose(*rcfileP)) {
          msg= Tcl_PosixError(interp);
          Tcl_AppendResult(interp, "couldn't write newly-created .groggsrc: ",
                           msg, (char*)0);
          ckfree(bp); return TCL_ERROR;
        }
        r= Tcl_SetVar(interp,argv[2],
                      "warning: you have no .groggsrc - I shall create one for you. "
                      "It will be set up to use the GUEST account.  You must register "
                      "a personal userid in order to be able to post.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) { ckfree(bp); return r; }
        *rcfileP= fopen(bp,"r");
        if (!*rcfileP) {
          msg= Tcl_PosixError(interp);
          Tcl_AppendResult(interp, "couldn't re-read newly-created .groggsrc: ",
                           msg, (char*)0);
          ckfree(bp); return TCL_ERROR;
        }
      } else {
        *secretfileP= 0; ckfree(bp); return TCL_OK;
      }
    } else {
      msg= Tcl_PosixError(interp);
      Tcl_AppendResult(interp, "failed to open .groggsrc: ", msg, (char*)0);
      ckfree(bp); return TCL_ERROR;
    }
  }
  strcpy(bp+l, ".groggsecret");
  *secretfileP= fopen(bp,"r");
  ckfree(bp);
  if (!*secretfileP && errno != ENOENT) {
    msg= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "failed to open .groggsecret: ", msg, (char*)0);
    fclose(*rcfileP); return TCL_ERROR;
  }
  if (fstat(fileno(*secretfileP,&stab))) {
    msg= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "failed to fstat .groggsecret: ", msg, (char*)0);
    fclose(*secretfileP); fclose(*rcfileP); return TCL_ERROR;
  }
  if (stab.st_mode & (S_IRGRP|S_IXUSR|S_IROTH|S_IXOTH)) {
    interp->result=
      "your .groggsecret may be readable by other users; I refuse to continue";
    fclose(*secretfileP); fclose(*rcfileP); return TCL_ERROR;
  }
  return TCL_OK;
}

int Grc_ParseServerToList(Tcl_Interp *interp, char *warningsvar,
                          char *argstart, Tcl_DString *dsP) {
  int r,i;
  Tcl_Dstring hostportnick[3];
  
  r= Grc_ParseServer(interp,warningsvar,argstart,
                     hostportnick, hostportnick+1, hostportnick+2);
  if (r != TCL_OK) return r;
  Tcl_DStringInit(dsP);
  for (i=0; i<3; i++) {
    Tcl_DStringAppendElement(dsP,Tcl_DStringValue(hostportnick+i));
    Tcl_DStringFree(hostportnick+i);
  }
  return TCL_OK;
}

int Groggsrc_ServerCmd(ClientData clientData, Tcl_Interp *interp,
                       int argc, char **argv) {
  static int id= 0;
  int process= 1;
  Grc_Server *servP;
  Grc_ItemInfo *itemP;
  FILE *files[];
  char *host, *port, **pp;
  int r, i, fi, newentry;
  char *commandword, *argstart, *ep;
  Tcl_HashEntry *ientP;
  Tcl_DString ds;

  if (argc != 5) { interp->result= "wrong # args"; return TCL_ERROR; }
  
  servP= (Grc_Server*)ckalloc(sizeof(Grc_Server));
  
  Tcl_DStringInit(&servP->itemtext);
  servP->id= id++;
  servP->userid[0]= 0;
  servP->md5secretvalid= 0;
  servP->seenallbeforevalid= 0;
  for (i=0; i<3; i++) Tcl_DStringInit(serverids+i);
  for (i=0; i<6; i++) Tcl_DStringInit(strings+i);
  Tcl_InitHashTable(&servP->items, TCL_STRING_KEYS);

  r= Grc_OpenFiles(interp,argv[1],argv[2], files,files+1,1);
  if (r != TCL_OK) return r;

  for (fi=0; fi<2; fi++) {
    r= Grc_ReadLine(interp,files[fi],&commandword,&argstart);
    if (r != TCL_OK) goto x_abortread;
    if (!commandword) break;
    if (!strcmp(commandword,"default-server")) {
      continue;
    } else if (!strcmp(commandword,"server")) {
      r= Grc_ParseServer(interp,argv[2], argstart, &host,&port);
      process= !strcmp(host,argv[3]) && !strcmp(port,argv[4]);
      continue;
    }
    if (!process) continue;
    if (!strcmp(commandword,"md5-secret")) {
      if (!fi) {
        interp->result= "md5-secret in the .groggsrc instead of the .groggsecret;"
          " I refuse to continue.";
        r= TCL_ERROR;  goto x_abortread;
      }
      servP->md5secret= strtoul(argstart,&ep,16);
      if (*ep) {
        r= Tcl_SetVar(interp,argv[2],
                      "md5-secret command not followed by hex digits.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) goto x_abortread;
        servP->md5secretvalid= 0;
      } else {
        servP->md5secretvalid= 1;
      }
      continue;
    } else if (!strcmp(commandword,"user")) {
      if (strlen(argstart) <= USERID_MAXLEN) {
        strcpy(servP->userid);
      } else {
        r= Tcl_SetVar(interp,argv[2],
                      "user command followed by username that is too long.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) goto x_abortread;
      }
    } else if (!strcmp(commandword,"seen-all-before")) {
      if (servP->seenallbeforevalid) {
        r= Tcl_SetVar(interp,argv[2],
                      "two seen-all-before commands.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) goto x_abortread;
      } else {
        servP->seenallbefore= strtoul(argstart,&ep,16);
        if (*ep) {
          r= Tcl_SetVar(interp,argv[2],
                        "seenallbefore command not followed by hex digits.\n"
                        TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
          if (r != TCL_OK) goto x_abortread;
          servP->seenallbeforevalid= 0;
        } else {
          servP->seenallbeforevalid= 1;
        }
      }
    } else if (!strcmp(commandword,"item")) {
      itemP= ckalloc(sizeof(Grc_ItemInfo));
      itemP->lastseenvalid= 0;
      itemP->junklevel= 0;
      p= argstart;
      q= strchr(argstart," ");
      if (q) { *q++= 0; qq= q; } else { qq= ""; }
      if (q != p+8) {
        r= Tcl_SetVar(interp,argv[2],
                      "item command not followed by itemid.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) goto x_abortread;
        ckfree(itemP); continue;
      }
      ip= p;
      if (*qq && !strncmp(qq,"lastseen ",9)) {
        itemP->lastseen= strtoul(qq+9,&ep,16);
        if (ep != qq+17 || (*ep != ' ' && *ep)) {
          r= Tcl_SetVar(interp,argv[2],
                        "item lastseen clause doesn't contain hex.\n"
                        TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
          if (r != TCL_OK) goto x_abortread;
          ckfree(itemP); continue;
        }
        itemP->lastseenvalid= itemP->lastseen;
        qq= ep;
        while (isspace(*qq)) qq++;
      }
      if (*qq) {
        if (!strcmp(qq,"junk")) {
          itemP->junklevel= 1;
        } else if (!strcmp(qq,"kill")) {
          itemP->junklevel= 1;
        } else {
          r= Tcl_SetVar(interp,argv[2],
                        "item command contains garbage.\n"
                        TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
          if (r != TCL_OK) goto x_abortread;
          ckfree(itemP); continue;
        }
      }
      ientP= Tcl_CreateHashEntry(&servP->items, ip, &newentry);
      if (!newentry) {
        r= Tcl_SetVar(interp,argv[2],
                      "two item commands for one item.\n"
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        if (r != TCL_OK) goto x_abortread;
        ckfree(itemP); continue;
      }
      Tcl_SetHashValue(ientP,itemP);
    } else {
      for (i=0, pp= stringnames;
           *pp && strcmp(*pp,commandword);
           pp++, i++);
      if (*pp) {
        Tcl_DStringFree(servP->strings+i);
        Tcl_DStringAppend(servP->strings+i, argstart, -1);
      } else {
        Tcl_DStringInit(&ds);
        Tcl_DStringAppend(&ds,"warning: unknown command `",-1);
        Tcl_DStringAppend(&ds,commandword);
        Tcl_DStringAppend(&ds,"' in .groggsrc/.groggsecret ignored.\n",-1);
        r= Tcl_SetVar(interp,argv[2], Tcl_DStringValue(&ds),
                      TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
        Tcl_DStringFree(&ds);
        if (r != TCL_OK) goto x_abortread;
      }
    }
  }
  for (fi=0; fi<2; fi++) fclose(files+fi);
  p= Grc_CommandName(servP);
  Tcl_CreateCommand(interp, p, Grc_ServerCmd,
                    (ClientData)servP, Grc_DeleteCallback);
  strcpy(interp->result, p);
  return TCL_OK;
  
 x_abortread:
  for (fi=0; fi<2; fi++) fclose(files[fi]);
  Grc_DestroyServer(&servP); return r;
}

int Grc_ListServersCmd(ClientData clientData, Tcl_Interp *interp,
                       int argc, char **argv) {
  FILE *rcfile, *secretfile;
  int r, had;
  char *argstart, *commandword;
  Tcl_DString overlist, elem;
  
  if (argc != 3) { interp->result= "wrong # args"; return TCL_ERROR; }
  
  r= Grc_OpenFiles(interp,argv[1],argv[2], &rcfile,&secretfile,0);
  if (r != TCL_OK) return r;

  Tcl_DStringInit(&overlist);
  had= 0;
  if (rcfile) {
    for (;;) {
      r= Grc_ReadLine(interp,rcfile,&commandword,&argstart);
      if (r != TCL_OK) { Tcl_DStringFree(&overlist); return r; }
      if (!commandword) break;
      if (!strcmp(commandword,"default-server")) {
        if (had) {
          r= Tcl_SetVar(interp,argv[2],
                        "warning: default-server in .groggsrc after"
                        " server or default-server\n",
                        TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
          if (r != TCL_OK) { Tcl_DStringFree(&overlist); return r; }
          continue;
        }
        r= Grc_ParseServerToList(interp,argv[2], argstart,&elem);
        if (r != TCL_OK) { Tcl_DStringFree(&overlist); return r; }
        Tcl_DStringAppendElement(&overlist,Tcl_DStringValue(&elem));
        Tcl_DStringFree(&elem);
        had++;
      } else if (!strcmp(commandword,"server")) {
        if (!haddefault) {
          Tcl_DStringAppendElement(&overlist,"");
          had++;
        }
        r= Grc_ParseServerToList(interp,argv[2], argstart,&elem);
        Tcl_DStringAppendElement(&overlist,Tcl_DStringValue(&elem));
        Tcl_DStringFree(&elem);
        had++;
      }
    }
  }
  if (!had) {
    r= Grc_ParseServerToList(interp,argv[2],"", &elem);
    Tcl_DStringAppendElement(&overlist,Tcl_DStringValue(&elem));
    Tcl_DStringAppendElement(&overlist,Tcl_DStringValue(&elem));
    Tcl_DStringFree(&elem);
  } else if (had==1) {
    r= Tcl_SetVar(interp,argv[2],
                  "warning: default-server in .groggsrc but"
                  " no server line appears later\n",
                  TCL_LEAVE_ERR_MSG|TCL_APPEND_VALUE);
    if (r != TCL_OK) { Tcl_DStringFree(&overlist); return r; }
    continue;
  }
  Tcl_DStringResult(interp,&overlist);
  return TCL_OK;
}

int Groggsrc_TclInit(Tcl_Interp *interp) {
  Tcl_CreateCommand(interp, "groggsrc-listservers",
                    Grc_ListServersCmd, (ClientData)0, (void(*)())0);
  Tcl_CreateCommand(interp, "groggsrc-server",
                    Grc_ServerCmd, (ClientData)0, (void(*)())0);
  return TCL_OK;
}
